{
{- 
   CodeGen.y  by Akimasa Morihata

   Parser for Optimal Path Query System (Happy File) 

   You can use two atomic predicates "from" and "to": 
       from(e,v)  =  in(e) == v
       to(e,v)  =  out(e) == v
   In addition, you can use an atomic predicate "train", 
   which is practically nonsense:
       train(e) == odd(in(e)) && odd(out(e)) 
   You can use a weight function "w".
-}

module Main where
import CodeGenMain
import Data.Char
import System.IO
import System
import Debug.Trace
}

%name parse
%tokentype { Token }
%error { parseError }

%token minimize        { TokenMinimize }
       "s.t."          { TokenST }
       where           { TokenWhere }
       id              { TokenID $$ }
       bool            { TokenBool $$ }
       '&&'            { TokenAnd }
       '||'            { TokenOr }
       '!'             { TokenNeg }
       int             { TokenInt $$ }
       if              { TokenIf } 
       then            { TokenThen }
       else            { TokenElse }
       min             { TokenMin }
       max             { TokenMax }
       '='             { TokenEq }
       "<="            { TokenLeq }
       '<'             { TokenLe }
       ">="            { TokenGeq }
       '>'             { TokenGe }
       '+'             { TokenPlus }
       '*'             { TokenTimes }
       '('             { TokenOB }
       ')'             { TokenCB }
       ','             { TokenComma }
       ';'             { TokenSemicolon }
       '.'             { TokenDot }

%% 

Prog : minimize id '(' id ')' "s.t." BExp where Decls 
        { ($2,$7 ($4,bot ""), $9) } 

Decls : BFunDecl Decls { $1 : $2 } 
      | IFunDecl Decls { $1 : $2 }
      | {- empty -}    { [] }

BFunDecl : id'('')' '=' BExp ';' id'('id'.'id')' '=' BExp ';'
           { if $1 == $7 
             then BoolFunc $1 ($5 (bot "", bot "")) ($14 ($9,$11)) 
             else bot "bfd"}  
         | id'('id'.'id')' '=' BExp ';' id'('')' '=' BExp ';' 
           { if $1 == $10 
             then BoolFunc $1 ($14 (bot "",bot "")) ($8 ($3,$5))
             else bot "bfd" }  

IFunDecl : id'('')' '=' IExp ';' id'('id'.'id')' '=' IExp ';'
           { if $1 == $7 
             then IntFunc $1 ($5 (bot "", bot "")) ($14 ($9,$11)) 
             else bot "ifd" }  
         | id'('id'.'id')' '=' IExp ';' id'('')' '=' IExp ';' 
           { if $1 == $10 
             then IntFunc $1 ($14 (bot "",bot "")) ($8 ($3,$5))
             else bot "ifd"}  

BExp : BExp '||' BExp0
       { \p -> ($1 p) :|: ($3 p) }
     | BExp0
       { $1 }

BExp0 : BExp0 '&&' BExp1
        { \p -> ($1 p) :&: ($3 p)}
      | BExp1
        { $1 }

BExp1 : bool 
        { \p -> CBool $1 }
      | '!' BExp1
        { \p -> Neg ($2 p) }
      | id '(' id ')'
        { \(x,e) -> if $3 == x
                       then BFunCall $1 
	                   else if $3 == e && $1 == "train" 
                            then AP "(e.in%2 && e.out%2)"
                            else error ("unknown atomic predicate:" ++ $1) }
      | id '(' id ',' int ')'
        { \(x,e) -> if $3 == e 
                    then if $1 == "from" then AP ("e.in==" ++ show $5) 
                 	 else if $1 == "to" then AP ("e.out==" ++ show $5) 
                              else error ("unknown atomic predicate:" ++ $1) 
                    else bot "apc" } 
      | '(' BExp ')'
        { $2 }
      | InEq
        { $1 } 

InEq : IExp '=' IExp
       { \p -> equiv ($1 p) ($3 p) } 
     | IExp '<' IExp
       { \p -> lt ($1 p) ($3 p) } 
     | IExp "<=" IExp 
       { \p -> leq ($1 p) ($3 p) } 
     | IExp '>' IExp
       { \p -> gt ($1 p) ($3 p) } 
     | IExp ">=" IExp
       { \p -> geq ($1 p) ($3 p) } 

IExp : if BExp then IExp else IExp
       { \p -> IfExp ($2 p) ($4 p) ($6 p) }
     | Add
       { $1 }

Add : Add '+' Mult
      { \p -> addExp ($1 p) ($3 p) } 
    | Mult
      { $1 }

Mult : Mult '*' Prim
       { \p -> multExp ($1 p) ($3 p) } 
     | Prim
       { $1 }

Prim : id '(' id ')'
       { \(x,e) -> if $3 == x 
                   then IFunCall $1 
                   else if $3 == e 
                        then if $1 == "w" 
                             then Weight "w"
                             else error "unknown weight function"
                        else bot "ifc" } 
     | int
       { \p -> CInt $1 } 
     | max '(' IExp ',' IExp ')'
       { \p -> maxExp ($3 p) ($5 p) } 
     | min '(' IExp ',' IExp ')'
       { \p -> minExp ($3 p) ($5 p) } 
     | '(' IExp ')'
       { $2 }

{
  main = do
         args <- getArgs 
         if args == []  
            then putStr "usage: ./OP <spec-file>"
		 else readFile (args!!0) >>= (generate . parse . lexer)

parseError ts = error ("Parse error: " ++ show ts)

bot = \s -> error ("parse error: "++s)

data Token = TokenMinimize
           | TokenST
           | TokenWhere
           | TokenID String
           | TokenBool Bool
           | TokenAnd
           | TokenOr
           | TokenNeg
           | TokenInt Int
           | TokenMin
           | TokenMax
           | TokenEq
           | TokenLeq
           | TokenLe 
           | TokenGeq 
           | TokenGe
           | TokenPlus
           | TokenTimes
           | TokenOB
           | TokenCB
           | TokenOBB
           | TokenCBB
           | TokenComma
           | TokenSemicolon
           | TokenDot
           | TokenIf
           | TokenThen
           | TokenElse
           deriving (Show,Eq)

lexer [] = []
lexer ('/':'/':cs) = skip1 cs
lexer ('/':'*':cs) = skip2 cs
lexer ('s':'.':'t':'.':cs) = TokenST : lexer cs
lexer ('&':'&':cs) = TokenAnd : lexer cs
lexer ('|':'|':cs) = TokenOr : lexer cs
lexer ('!':cs) = TokenNeg : lexer cs
lexer ('=':cs) = TokenEq : lexer cs
lexer ('<':'=':cs) = TokenLeq : lexer cs
lexer ('<':cs) = TokenLe : lexer cs
lexer ('>':'=':cs) = TokenGeq : lexer cs
lexer ('>':cs) = TokenGe : lexer cs
lexer ('+':cs) = TokenPlus : lexer cs
lexer ('*':cs) = TokenTimes : lexer cs
lexer ('(':cs) = TokenOB : lexer cs
lexer (')':cs) = TokenCB : lexer cs
lexer ('[':cs) = TokenOBB : lexer cs
lexer (']':cs) = TokenCBB : lexer cs
lexer (',':cs) = TokenComma : lexer cs
lexer (';':cs) = TokenSemicolon : lexer cs
lexer ('.':cs) = TokenDot : lexer cs
lexer (c:cs) | isSpace c = lexer cs
             | isAlpha c = lexAlpha (c:cs)
             | isDigit c = lexNum (c:cs)

lexAlpha cs =
   case span isIDChar cs of
      ("minimize",rest) -> TokenMinimize : lexer rest
      ("where",rest)  -> TokenWhere : lexer rest
      ("true",rest)  -> TokenBool True : lexer rest
      ("false",rest)  -> TokenBool False : lexer rest
      ("if",rest)  -> TokenIf : lexer rest
      ("min",rest)  -> TokenMin : lexer rest
      ("max",rest)  -> TokenMax : lexer rest
      ("then",rest)  -> TokenThen : lexer rest
      ("else",rest)  -> TokenElse : lexer rest
      (var,rest)   -> TokenID var : lexer rest

isIDChar c = isAlpha c || isDigit c || c == '_' || c == '-'

lexNum cs = TokenInt (read num) : lexer rest
      where (num,rest) = span isDigit cs

skip1 [] = []
skip1 ('\n':cs) = lexer cs
skip1 (_:cs) = skip1 cs
 
skip2 [] = []
skip2 ('*':'/':cs) = lexer cs
skip2 (_:cs) = skip2 cs

}
