{- 
   CodeGenMain.hs  by Akimasa Morihata

   Code generator:
   analysing query description and generating C++ code.
 
   The generated code consists of the following:
   - definition of a struct "val", which retains the informations of the path and used to compute "state(x)" and so on. 
   - definition of a hash function, which is used for hash tables.
   - definition of a function object that checks equivalence of paths on "state(x)".
   - definition of the constraint the solutions should satisfy.
   - definition of a predicate that checks wheter the path never yeilds a feasible path. (optimization)
-}

module CodeGenMain where

import Data.List
import Data.Either
import Debug.Trace
import System.IO


-- ===================  Definition of AST  =================== --


type Prog = (String,BoolExp, [FunDecl])
-- specification. "(f, p, decls)" means: 
-- "Find a path of minimum_f value that satisfies p, where decls". 

data FunDecl = BoolFunc String BoolExp BoolExp 
             | IntFunc String IntExp IntExp
instance Show (FunDecl) where show = showDecl

data BoolExp = Neg BoolExp                -- logical negation 
             | BoolExp :&: BoolExp        -- logical conjunction
             | BoolExp :|: BoolExp        -- logical disjunction
             | BFunCall String            -- function call
             | AP String                  -- atomic predicate
             | BinEq BinEq IntExp IntExp  -- inequality 
             | CBool Bool                 -- constant boolean value
               deriving Eq
{-
  Note:
  We can write ANY C++ code for AP, and the sysytem just output it.
-}

data BinEq = LEQ | EQIV
             deriving Eq

data IntExp = BinOp BinOp IntExp IntExp    -- binary operator
            | IfExp BoolExp IntExp IntExp  -- "if" expression
            | IFunCall String              -- function call 
            | Weight String                -- weight of the edge
            | CInt Int                     -- constant non-negative integer
              deriving Eq

data BinOp = ADD | TIMES | MAX | MIN
             deriving Eq


-- for pretty printing  
instance Show (BoolExp) where show = showBoolExp
showBoolExp e = case e of  
                  Neg e -> "!(" ++ show e ++")"
                  e1 :|: e2 -> "(" ++ show e1 ++" || " ++ show e2 ++ ")"
                  e1 :&: e2 -> "(" ++ show e1 ++" && " ++ show e2 ++ ")"
                  BFunCall s -> s ++"(x)"
                  AP s -> s
                  BinEq b e n -> "(" ++ show e ++ show b ++ show n ++ ")"
                  CBool True -> "T"
                  CBool False -> "F"

instance Show (IntExp) where show = showIntExp
showIntExp e = case e of 
                 BinOp b e1 e2 -> showBinOp b (show e1) (show e2)
                 IfExp p e1 e2 -> "(if" ++ show p ++" then " ++ 
                                  show e1 ++ " else " ++ show e2 ++ ")"
                 IFunCall s -> s ++"(x)"
                 Weight s -> "e."++ s
                 CInt n  -> show n

showBinOp e s1 s2 = case e of ADD -> "(" ++ s1 ++ "+" ++ s2 ++ ")"
                              TIMES -> "(" ++ s1 ++ "*" ++ s2 ++ ")"
                              MIN -> "std::min(" ++ s1 ++ "," ++ s2 ++")"
                              MAX -> "std::max(" ++ s1 ++ "," ++ s2 ++")"
instance Show (BinEq) where show = showBinEq

showBinEq e = case e of LEQ -> "<="
                        EQIV -> "=="

-- showProg is the pretty printer
showProg (f, p, fs) = 
    "\nmin. " ++ f ++ "(x)\ns.t. " ++ show p ++ "\n  where \n" ++ 
                foldr1 (\l r-> l ++ r) (map show fs) ++"\n"

showDecl (BoolFunc s e1 e2 ) = 
    "  " ++ s ++ "([]) = "  ++ show e1 ++ ";\n  " ++ s ++ "(xe) = " ++ show e2 ++ ";\n"
showDecl (IntFunc s  e1 e2 ) = 
    "  " ++ s ++ "([]) = "  ++ show e1 ++ ";\n  " ++ s ++ "(xe) = " ++ show e2 ++ ";\n"


-- utilities

addExp e1 e2 = BinOp ADD e1 e2  
multExp e1 e2 = BinOp TIMES e1 e2  
maxExp e1 e2 = BinOp MAX e1 e2  
minExp e1 e2 = BinOp MIN e1 e2  

lt e1 (CInt n) = BinEq LEQ e1 (CInt $ n-1)
leq e1 e2 = BinEq LEQ e1 e2
equiv e1 e2 = BinEq EQIV e1 e2
gt e1 e2 = Neg (BinEq LEQ e1 e2)
geq e1 e2 = Neg $ lt e1 e2




-- ===================  Translater =================== --

-- "generate" is the main function, which outputs C++ codes "val.cpp"
generate :: Prog -> IO ()
generate prog = generateMain prog "val.cpp" 

generateMain :: Prog -> String -> IO ()
generateMain prog filename = do
  fp <- openFile filename WriteMode
  let fputs = hPutStr fp
  fputs "//this program is generated by the system!\n\n"
  fputs "#include \"node_edge.h\"\n\n"
  fputs $ "#define PRIME_FOR_HASH " ++ show prime1 ++"\n"
  fputs $ "#define PRIME_FOR_HASH_ " ++ show prime2 ++"\n\n"
  fputs $ showValDef prog
  fputs $ showHash prog
  fputs $ showStateEq prog
  fputs $ showConstraint prog
  fputs $ showRemoveState prog
  hClose fp


-- for printing C++ expressions

transIntExp val e = 
  case e of 
    BinOp b e1 e2 -> showBinOp b (transIntExp val e1) (transIntExp val e2)
    IfExp p e1 e2 -> "(" ++ transBoolExp val p ++" ? " ++ 
                     transIntExp val e1 ++ " : " ++ 
                     transIntExp val e2 ++")" 
    IFunCall s -> val++s
    Weight s -> "e."++s
    CInt n  -> show n

transBoolExp val e = 
  case e of 
    Neg e -> "!" ++ transBoolExp val e ++ ""
    e1 :&: e2 -> "("++transBoolExp val e1++" && "++transBoolExp val e2++")"
    e1 :|: e2 -> "("++transBoolExp val e1++" || "++transBoolExp val e2++")"
    BFunCall s -> val ++ s
    AP s -> s
    BinEq b e1 e2 -> "("++transIntExp val e1++show b++transIntExp val e2++")" 
    CBool True -> "true"
    CBool False -> "false"



funDecls (e,p,decls) = map aux1 decls 
    where aux1 (IntFunc p e1 e2) = Left p
          aux1 (BoolFunc p e1 e2) = Right p


-- printing the definition of the memo table. 

showValDef p@(ew,_,ds) = 
    "struct val {\n"++valDec++init++update++ 
                    showObjective p++showNinf p++"};\n\n"
    where
      ls = map (either ("int "++) ("bool "++)) $ funDecls p
      valDec = catMap (\a->"\t" ++ a ++ ";\n") ls ++ "\tnode n;\n"
      init = "\tval() {\n\t\tn = -1;\n"++catMap aux2 ds++"\t}\n"
      update = "\tval(val v,edge e) {\n" ++catMap aux3 ds++edef++"\t}\n"
      edef = "\t\tn = e.out;\n"
      aux2 (BoolFunc p e1 _) = "\t\t"++p++" = "++transBoolExp "p.v." e1++";\n"
      aux2 (IntFunc p e1 _) = "\t\t"++p++" = "++transIntExp "p.v." e1 ++";\n"
      aux3 (BoolFunc p _ e2) = "\t\t"++p++" = "++transBoolExp "v." e2++";\n"
      aux3 (IntFunc p _ e2) = "\t\t"++p++" = "++transIntExp "v." e2++";\n"


-- gather all inequalities in the program

inEqs :: Prog -> [(IntExp, Int)]
inEqs (e,p,ds) = sortBy (compare `onf` snd) $ concat $ auxB p : map aux ds
    where aux (BoolFunc s e1 e2) = auxB e1 ++ auxB e2
          aux (IntFunc s e1 e2) = auxI e1 ++ auxI e2
          auxI (BinOp b e1 e2) = auxI e1 ++ auxI e2
          auxI (IfExp p e1 e2) =  auxB p ++ auxI e1 ++ auxI e2 
          auxI _ = []
          auxB (Neg e) = auxB e
          auxB (e1 :&: e2) = auxB e1 ++ auxB e2 
          auxB (e1 :|: e2) = auxB e1 ++ auxB e2 
          auxB (BinEq b e (CInt n)) = (e,n) : auxI e 
          auxB _ = []

     
-- generate the definition of the "state" function 

states :: Prog -> [Either (String,Int) String] 
states prog@(e,p,ds) = 
    let preds = boolFuncs ds
        boolFuncs ((BoolFunc s e1 e2):x) = Right s : boolFuncs x
        boolFuncs ((IntFunc s e1 e2):x) = boolFuncs x
        boolFuncs [] = []
        dict = intBody ds
        intBody ((BoolFunc s e1 e2):x) = intBody x
        intBody ((IntFunc s e1 e2):x) = (s,e2) : intBody x
        intBody [] = []
    in preds ++ (map Left . fst $ foldr aux ([],dict) (inEqs prog))
 where aux (BinOp b e1 e2,n) fsd = aux (e2,n) (aux (e1,n) fsd)
       aux (IfExp p e1 e2,n) fsd = aux (e2,n) (aux (e1,n) fsd)
       aux (IFunCall s,n) fsd@(fs,d) = 
           case lookup s d of 
                  Nothing -> fsd
                  Just e -> aux (e,n) ((s,n):fs, filter ((/=s).fst) d)
       aux _ fsd = fsd


-- printing the definition of objective function
showObjective (s,_,_) = "\tint weight() const { return "++ s ++"; }\n"

-- printing the definition of "INT_MIN", which is used in the heap.
showNinf (s,_,_) = "\tstatic val ninf() {\n\t\tval v;\n\t\tv." ++ s ++ 
                   " = INT_MIN;\n\t\treturn v;\n\t}\n"


-- printing the definition of the hash function.
showHash prog = 
    "struct val_hash_t {\n\tunsigned long operator()(const val &v) const {\n\t\tunsigned long long k = v.n;\n" ++ catMap (either f1 f2) (states prog) ++ 
    "\t\treturn (unsigned long)k;\n\t}\n} val_hash;\n\n" 
  where 
    f1 (p,k) = "\t\tk = (k * PRIME_FOR_HASH + ((v."++p++"<="++show k ++ ")?v."++p++":"++show (k+1)++")) % PRIME_FOR_HASH_;\n"
    f2 p = "\t\tk = (k * PRIME_FOR_HASH + v."++p++") % PRIME_FOR_HASH_;\n"


-- printing the definition of the function object to the check whether states of paths are the same 
showStateEq prog = 
  let ss = lHomo (\l r-> l++" && \n\t\t\t"++r) (either f1 f2) $ states prog
      f1 (p,k) = "((v."++p++"<="++show k++")?v."++p++":"++show (k+1)++
                 ") == ((w."++p++"<="++show k++")?w."++p++":"++show (k+1)++")"
      f2 p = "v."++p++" == "++"w."++p  
  in "struct val_eq_t {\n\tbool operator()(const val &v, const val &w) const {\n\t\treturn (\n\t\t\tv.n == w.n && \n\t\t\t" ++ ss ++"\n\t\t);\n\t}\n} val_eq;\n\n"


-- printing the constraint that solutions should satisfy.
showConstraint (_,p,_) = "inline bool constraint( val v ) {\n\treturn " 
                         ++ transBoolExp "v." p ++";\n}\n"


-- for optimization 

showRemoveState prog = 
    let sts = states prog
        un = map (zip sts) $ removeState prog
        aux (Left (s,k), b) = if b then "(a."++s ++ "<=" ++ show k ++")"
                          else "(a."++ s ++ ">" ++ show k ++")"
        aux (Right s, b) = if b then "a."++s else "(!" ++ "a."++s++")"
    in "inline bool unnecessary( val a ) {\n \treturn (\n" ++ 
       lHomo (\a r-> a++ " || \n"++r) 
                 ( ("\t\t("++).(++")").lHomo (\a r->a++" && "++r) aux) un ++ 
                 "\n\t);\n}\n"

removeState prog = unnecessary prog (makeTransGraph prog)

-- generate the automaton that corresponds to the "state" function.
makeTransGraph prog@(_,p,decls) = 
    let sts = allStates $ map (either (Left . fst) Right) $ states prog
        ds = map aux decls
    in zip (map (map snd) sts) (map (nextState ds) sts)  
    where aux (BoolFunc s _ e) = (s, Right e)
          aux (IntFunc s _ e) = (s, Left e)

-- generate all states of the automaton
allStates = foldr (\s r->map ((s,True):) r ++ map ((s,False):) r) [[]] 

-- obtain the transition
nextState decls st = foldr (\x rs->[a:r|a<-x,r<-rs]) [[]] $ map aux st 
    where aux (f@(Left s), b) = let Left e = lookupFJ s decls
                                in if not b && increase s st e 
                                   then [False] else [True,False]
          aux (p@(Right s), b) = let Right e = lookupFJ s decls in eval e st

-- Evaluate the expressions to obtain the transition
eval :: BoolExp -> [(Either String String, Bool)] -> [Bool]
eval e st = 
  case e of 
    Neg e -> map not $ eval e st
    e1 :&: e2 -> let x = eval e1 st; y = eval e2 st 
                 in (if (elem True x)&&(elem True y) then [True] else []) ++ 
                    (if (elem False x)||(elem False y) then [False] else [])
    e1 :|: e2 -> let x = eval e1 st; y = eval e2 st 
                 in (if (elem True x)||(elem True y) then [True] else []) ++ 
                    (if (elem False x)&&(elem False y) then [False] else [])
    BFunCall s -> [lookupFJ (Right s) st]
    AP _ -> [True,False]
    BinEq b e _ -> let fs = funCalls e st
                   in if all (\f-> lookupFJ (Left f) st) fs 
                      then [True,False] else [False] 
    CBool b -> [b]


-- Checking whether a integer-valued function is increasing
-- (Since integer values are abstracted out (to -infty or +infty), checking incleasingness is important)
increase f st e = aux e 
  where aux e = case e of CInt i -> False
                          BinOp MIN e1 e2 -> aux e1 && aux e2
                          BinOp TIMES e1 e2 -> aux e1 && aux e2
                          BinOp ADD e1 e2 -> aux e1 || aux e2
                          BinOp MAX e1 e2 -> aux e1 || aux e2
                          IfExp p e1 e2 -> let r = eval p st
                                           in (if True `elem` r then aux e1 
                                               else True) &&
                                              (if False `elem` r then aux e2 
                                               else True)
                          IFunCall s -> if s == f then True else False
                          Weight s -> False

-- gather all function calls 
funCalls :: IntExp -> [(Either String String, Bool)] -> [String]
funCalls e st = 
    case e of 
      BinOp b e1 e2 -> funCalls e1 st `union` funCalls e2 st
      IfExp p e1 e2 -> let tf = eval p st
                       in (if True `elem` tf then funCalls e1 st else []) ++  
                              (if False `elem` tf then funCalls e2 st else [])
      IFunCall s ->[s]
      _ -> []

-- check unnecessary states by doing BSF from acceptable states
unnecessary prog@(_,p,_) trans = 
    let sts = allStates $ map (either (Left . fst) Right)  $ states prog 
        acc = filter (accept p) sts  
        rest = sts \\ acc 
    in aux trans (map (map snd) acc) (map (map snd) rest)
    where aux _ _ [] = []
          aux _ [] rest = rest
          aux trans acc rest = 
              let prevs = map fst $ filter (overlap acc . snd) trans
                  acc' = rest `intersect` prevs  
                  rest' = rest \\ acc'
              in aux trans acc' rest'
          accept p st = elem True (eval p st) 
          overlap x y = intersect x y /= []


-- primes (used for hash functions)
prime1 = 402653189
prime2 = 1610612741


--utilities

(*) `onf` f = \x y -> f x * f y

catMap f = concat . map f
lHomo op f = foldr1 op . map f
lookupFJ k l = case lookup k l of Just x -> x
                                  Nothing -> error (show k)


